/***************************************************************************************************/
/*                                                                                                 */
/* file:          Controller.cpp                                                                   */
/*                                                                                                 */
/* source:        2015-2106, written by Adrian Kundert                                             */
/*                                                                                                 */
/* description:   The main class for the Self-Balanced PT project                                  */
/*                                                                                                 */
/***************************************************************************************************/

#include "Controller.h"

#define SABERTOOTH_ADDRESS 128

#define SABERTOOTH_DRIVE_FWD_MOTOR1 0
#define SABERTOOTH_DRIVE_REV_MOTOR1 1
#define SABERTOOTH_DRIVE_FWD_MOTOR2 4
#define SABERTOOTH_DRIVE_REV_MOTOR2 5

//---------------------------------debugging variable-----------------------------------------------//
unsigned long printdelay = 0;
bool overrun = false;
       
Controller::Controller() :
   mpuTilt(0x68),   // default I2C address AD0 low for the GIY board #2
   mpuSteer(0x69)     // I2C address AD0 high for the GIY board #3
{
    //initialize the MPU objects
    mpuTilt.setAccelOffset(377, -42, -1126);    mpuTilt.setGyroOffset(-581, 34, -248);
    mpuSteer.setAccelOffset(40,-89, -318);      mpuSteer.setGyroOffset(-132, 121, -97);

    isrCnt = 0;
    MPUinterruptFlag = false;
    
    State = Idle;
    ReadySwCnt = 0; ReadySw = false;
    Kp = 3; Ki = 100; Kd = 5;
    P = I = D = LastError = 0;
    BalancingCnt = 0;
    SpeedMotorLeft = SpeedMotorRight = DirMotorLeft = DirMotorRight = CmdSpeedMotor = 0;
    SteerAngle = 0;
}

void Controller::initialize() {

    mpuSteer.initialize(false, false); // no DMP, data read only by polling, calibration on
    mpuTilt.initialize(true, false);   // DMP activated for the MPU6050 Interrupt usage, calibration off    
}

void Controller::process() {

    // execute when at MPU interrupt
    if (MPUinterruptFlag == true) {
      // calculate the angle with MPU data
      mpuTilt.processData();
      mpuSteer.processData();
      MPUinterruptFlag = false;
      SteerAngle = mpuSteer.AngleAccel.y/25 + SteerAngle*24/25; // smooth the accel steer (63% @ 25*INT period = 0.5 sec, 90% @ 1 sec) 
      
      //PID controlled by y axis: since the desired angle should be 0, the angle is the error
      const byte AngleGain = 100;
      const int IntegratorLimit = 20000;
      long Error = (long)(AngleGain * -mpuTilt.Angle.x); // the angle is negativ when lean forward
      P = Kp * Error;
      I += (Ki * Error)/AngleGain;
      if (I > IntegratorLimit) I = IntegratorLimit; // limit integration accumulation
      if (I < -IntegratorLimit) I = -IntegratorLimit;
      D = Kd * (Error - LastError); LastError = Error;
      
      // handle the RUN switch
      if (digitalRead(READY_SW_PIN) == true) {
        if (ReadySwCnt < ReadySwCntMax) ReadySwCnt++;
        else ReadySw = true;
      }
      else {
        if (ReadySwCnt > 0) ReadySwCnt--;
        else ReadySw = false;
      }
  
      // FSM
      if (ReadySw == false) State = Idle;
      switch(State) {
        case Idle: //----------------------- initial state ----------------------------//
          P = I = D = 0; // clear the PID values when the state is not running
          if (ReadySw == true) State = StandingUp;        
          break;
        case StandingUp: //------- standing up the PT until within +-2° ---------------//
          I = D = 0;  //use the P only       
          if (abs(Error/AngleGain) < 2) State = Up;
          break;
        case Up:
          P = I = D = 0; // clear the PID values when the state is not running
          if (digitalRead(LOADED_PIN) == true) State = Balancing;
          break;
        case Balancing: //------- give a second to stabilize ---------------//
           if (++BalancingCnt >= 50) {
              State = Running; 
              BalancingCnt = 0;
           }
          break;
        case Running:
          if ((abs(SteerAngle) > 5) && (abs(Error/AngleGain) < 2) && ((SpeedMotorLeft < 5) || (SpeedMotorRight < 5))) State = Pivot;
          break;
        case Pivot:
          if (abs(SteerAngle) < 5) State = Running;
          if (abs(Error/AngleGain) > 3) State = Balancing;
      }
      long PID = (P+I+D)/AngleGain;
           
      // map it to the range of the sabertooth and send it
      char STcommandLeft = SABERTOOTH_DRIVE_FWD_MOTOR1;   // default forward
      char STcommandRight = SABERTOOTH_DRIVE_FWD_MOTOR2;   // default forward
      char STdataLeft, STdataRight;
      if (PID < 0) {
        // reverse direction for a negative value
        PID = -PID;
        STcommandLeft = SABERTOOTH_DRIVE_REV_MOTOR1;
        STcommandRight = SABERTOOTH_DRIVE_REV_MOTOR2;
      }
      if (PID > 127) PID = 127;  // limtation
      
      //reduction according to the steer (full left at +10.0°, full right at -10°)
      float steerRatio = abs(SteerAngle) / 10;
      if (steerRatio > 1.0) steerRatio = 1.0;
      char PIDscaled = (char)(PID * (1.0 - (1.0-PID/150.0)*steerRatio));
      if (SteerAngle > 0) {
        // steer to the left
        STdataRight = PID;
        if (State == Pivot) {
          STdataLeft = 0;
        }
        else {
          // left and right in the same direction
          STdataLeft = PIDscaled; 
        }
      }
      else {
        // steer to the right
        STdataLeft = PID;       
        if (State == Pivot) {
          STdataRight = 0;
        }
        else {
          STdataRight = PIDscaled; 
        }
      }

      // override when a cmd is active
      if ((CmdSpeedMotor != 0) && (ReadySw == true)) {
          STdataLeft = STdataRight = CmdSpeedMotor;
      }

      // send when speed value has changed
      if ((STdataLeft != SpeedMotorLeft) || (STcommandLeft != DirMotorLeft)) { 
        // left motor
        Serial.write(SABERTOOTH_ADDRESS);
        Serial.write(STcommandLeft);
        Serial.write(STdataLeft);
        char chksum = (SABERTOOTH_ADDRESS + STcommandLeft + STdataLeft) & 0x7f;
        Serial.write(chksum);
        DirMotorLeft = STcommandLeft;
        SpeedMotorLeft = STdataLeft;
      }
      if ((STdataRight != SpeedMotorRight) || (STcommandRight != DirMotorRight)) { 
        // right motor
        Serial.write(SABERTOOTH_ADDRESS);
        Serial.write(STcommandRight);
        Serial.write(STdataRight);
        char chksum = (SABERTOOTH_ADDRESS + STcommandRight + STdataRight) & 0x7f;
        Serial.write(chksum);
        DirMotorRight = STcommandRight;
        SpeedMotorRight = STdataRight;
      }
    }
     
    if (overrun == true) {
      overrun = false;
      rbuffer.write("!");    
      rbuffer.write((long)isrCnt);      
    }
        
    // periodic debug output
    const int PERIOD = 250;
    if (millis() - printdelay >= PERIOD) {
      printdelay += PERIOD;      
      rbuffer.write("\n");
      //rbuffer.write((long)isrCnt);
      rbuffer.write((int)(isrCnt&0xff));
      rbuffer.write("\t");
      
//      rbuffer.write(mpuSteer.GyroScaledValue.x,2,2);
//      rbuffer.write("\t");
//      rbuffer.write(mpuSteer.GyroScaledValue.y,2,2);
//      rbuffer.write("\t");
//      rbuffer.write(mpuSteer.GyroScaledValue.z,2,2);
//      rbuffer.write("\t");
//
//      rbuffer.write(mpuSteer.AccelScaledValue.x,2,2);
//      rbuffer.write("\t");
//      rbuffer.write(mpuSteer.AccelScaledValue.y,2,2);
//      rbuffer.write("\t");
//      rbuffer.write(mpuSteer.AccelScaledValue.z,2,2);
//      rbuffer.write("\t");
//
//      rbuffer.write(mpuSteer.AngleGyro.x,2,2);
//      rbuffer.write("\t");
//      rbuffer.write(mpuSteer.AngleGyro.y,2,2);
//      rbuffer.write("\t");
//      rbuffer.write(mpuSteer.AngleGyro.z,2,2);
//      rbuffer.write("\t");
//
//      rbuffer.write(mpuSteer.AngleAccel.x,2,2);
//      rbuffer.write("\t");
//      rbuffer.write(mpuSteer.AngleAccel.y,2,2);
//      rbuffer.write("\t");
//      rbuffer.write(mpuSteer.AngleAccel.z,2,2);
//      rbuffer.write("\t");
      rbuffer.write(SteerAngle,2,2);
      rbuffer.write("\t");

      rbuffer.write(mpuTilt.Angle.x,2,2);
      rbuffer.write("\t");
//      rbuffer.write(mpuTilt.Angle.y,2,2);
//      rbuffer.write("\t");
//      rbuffer.write(mpuTilt.Angle.z,2,2);
//      rbuffer.write("\t");

      rbuffer.write(P);
      rbuffer.write("\t");
      rbuffer.write(I);
      rbuffer.write("\t");
//      rbuffer.write(D);
//      rbuffer.write("\t");
//      rbuffer.write(Run);
//      rbuffer.write("\t");
//      rbuffer.write(State);
//      rbuffer.write("\t");
      rbuffer.write(SpeedMotorLeft);
      rbuffer.write("\t");
      rbuffer.write(SpeedMotorRight);
    }
}

// method called by MPU interrupt
void Controller::MPUinterrupt() {

    mpuTilt.getRawData();
    mpuTilt.resetFIFO(); // reset interrupt flag and flush the fifo content to permit further interrupts. The data is read from the registers.

    mpuSteer.getRawData();

    if (MPUinterruptFlag == true) overrun = true;
    MPUinterruptFlag = true;
    isrCnt++;
}

bool Controller::serialReadisCR() {
  bool retval = false;
  if (Serial.peek() == '\r') {
    Serial.read();  // extract the '\r' character  
    retval = true;
  }
  return retval;
}

void Controller::serialPrint() {
    char singleChar;
    if (rbuffer.getChar(&singleChar) == true) { 
      Serial.write(singleChar);
    }

    // check if the shortest command could be in the stream
    if (Serial.available() > 1) {

        singleChar = Serial.read();  // get the command
        switch (singleChar) {
        case '?':   // help command
          {
              if (serialReadisCR()) {
                  rbuffer.write("\n?\t\t- Help for the comands");
                  rbuffer.write("\np,i,d [value]   - Set/Get PID gain, value between 0 to 255");
                  rbuffer.write("\nl,r [value]   - Force the motor speed, value between -127 to 127");
                  rbuffer.write("\nv\t\t- Software version");                  
              }
              break;
          }
        case 'p':   // set/get PROPORTIONAL gain command
          {              
              byte gain = Kp;
              // check if a parameter is available or not
              bool isCR = serialReadisCR();
              if (isCR == false) {
                  gain = Serial.parseInt(); //parseInt() reads an ascii value until a non number (e.g. ',' or ' ' or '\n')     
                  isCR = serialReadisCR();
              }
              // once parameter extracted, verified CR
              if (isCR == true) {
                  Kp = gain;
                  rbuffer.write("\nKp = "); rbuffer.write(Kp);
              }
              break;
          }
        case 'i':   // set/get INTEGRAL gain command
          {
              byte gain = Ki;
              // check if a parameter is available or not
              bool isCR = serialReadisCR();
              if (isCR == false) {
                  gain = Serial.parseInt(); //parseInt() reads an ascii value until a non number (e.g. ',' or ' ' or '\n')     
                  isCR = serialReadisCR();
              }
              // once parameter extracted, verified CR
              if (isCR == true) {
                  Ki = gain;
                  rbuffer.write("\nKi = "); rbuffer.write(Ki);
              }
              break;
          }
        case 'd':   // set/get DERIVATIVE gain command
          {
              byte gain = Kd;
              // check if a parameter is available or not
              bool isCR = serialReadisCR();
              if (isCR == false) {
                  gain = Serial.parseInt(); //parseInt() reads an ascii value until a non number (e.g. ',' or ' ' or '\n')     
                  isCR = serialReadisCR();
              }
              // once parameter extracted, verified CR
              if (isCR == true) {
                  Kd = gain;
                  rbuffer.write("\nKd = "); rbuffer.write(Kd);
              }
              break;
          }
        case 'l':   // Force the motor speed
          {
              char motorSpeed;
              // check if a parameter is available or not
              bool isCR = serialReadisCR();
              if (isCR == false) {
                  motorSpeed = Serial.parseInt(); //parseInt() reads an ascii value until a non number (e.g. ',' or ' ' or '\n')     
                  isCR = serialReadisCR();
              }
              // once parameter extracted, verified CR
              if (isCR == true) {
                  CmdSpeedMotor = motorSpeed;
                  rbuffer.write("\nspeed = "); rbuffer.write(CmdSpeedMotor);
              }
              break;
          }
        case 'v':   // version command
          {
              if (Serial.peek() == '\r') {
                  Serial.read();  // extract the '\r' character
  
                  rbuffer.write("\nThe current software version is "); rbuffer.write(SwVersion);
              }
              break;
          }
        }
        rbuffer.write("\n\n");
    }
}


